@@ -1,12 +1,13 @@
-import torch
-import numpy as np
 from copy import deepcopy
+from typing import Any, Dict, Optional, Tuple, Union
+
+import numpy as np
+import torch
 from torch.distributions import Independent, Normal
-from typing import Any, Dict, Tuple, Union, Optional
 
-from tianshou.policy import DDPGPolicy
-from tianshou.exploration import BaseNoise
 from tianshou.data import Batch, ReplayBuffer, to_torch_as
+from tianshou.exploration import BaseNoise
+from tianshou.policy import DDPGPolicy
 
 
 class SACPolicy(DDPGPolicy):
@@ -15,32 +16,32 @@ class SACPolicy(DDPGPolicy):
     :param torch.nn.Module actor: the actor network following the rules in
         :class:`~tianshou.policy.BasePolicy`. (s -> logits)
     :param torch.optim.Optimizer actor_optim: the optimizer for actor network.
-    :param torch.nn.Module critic1: the first critic network. (s, a -> Q(s,
-        a))
+    :param torch.nn.Module critic1: the first critic network. (s, a -> Q(s, a))
     :param torch.optim.Optimizer critic1_optim: the optimizer for the first
         critic network.
-    :param torch.nn.Module critic2: the second critic network. (s, a -> Q(s,
-        a))
+    :param torch.nn.Module critic2: the second critic network. (s, a -> Q(s, a))
     :param torch.optim.Optimizer critic2_optim: the optimizer for the second
         critic network.
-    :param action_range: the action range (minimum, maximum).
-    :type action_range: Tuple[float, float]
-    :param float tau: param for soft update of the target network, defaults to
-        0.005.
-    :param float gamma: discount factor, in [0, 1], defaults to 0.99.
+    :param float tau: param for soft update of the target network. Default to 0.005.
+    :param float gamma: discount factor, in [0, 1]. Default to 0.99.
     :param (float, torch.Tensor, torch.optim.Optimizer) or float alpha: entropy
-        regularization coefficient, default to 0.2.
+        regularization coefficient. Default to 0.2.
         If a tuple (target_entropy, log_alpha, alpha_optim) is provided, then
-        alpha is automatatically tuned.
-    :param bool reward_normalization: normalize the reward to Normal(0, 1),
-        defaults to False.
-    :param bool ignore_done: ignore the done flag while training the policy,
-        defaults to False.
-    :param BaseNoise exploration_noise: add a noise to action for exploration,
-        defaults to None. This is useful when solving hard-exploration problem.
+        alpha is automatically tuned.
+    :param bool reward_normalization: normalize the reward to Normal(0, 1).
+        Default to False.
+    :param BaseNoise exploration_noise: add a noise to action for exploration.
+        Default to None. This is useful when solving hard-exploration problem.
     :param bool deterministic_eval: whether to use deterministic action (mean
-        of Gaussian policy) instead of stochastic action sampled by the policy,
-        defaults to True.
+        of Gaussian policy) instead of stochastic action sampled by the policy.
+        Default to True.
+    :param bool action_scaling: whether to map actions from range [-1, 1] to range
+        [action_spaces.low, action_spaces.high]. Default to True.
+    :param str action_bound_method: method to bound action to range [-1, 1], can be
+        either "clip" (for simply clipping the action) or empty string for no bounding.
+        Default to "clip".
+    :param Optional[gym.Space] action_space: env's action space, mandatory if you want
+        to use option "action_scaling" or "action_bound_method". Default to None.
 
     .. seealso::
 
@@ -56,22 +57,19 @@ class SACPolicy(DDPGPolicy):
         critic1_optim: torch.optim.Optimizer,
         critic2: torch.nn.Module,
         critic2_optim: torch.optim.Optimizer,
-        action_range: Tuple[float, float],
         tau: float = 0.005,
         gamma: float = 0.99,
-        alpha: Union[
-            float, Tuple[float, torch.Tensor, torch.optim.Optimizer]
-        ] = 0.2,
+        alpha: Union[float, Tuple[float, torch.Tensor, torch.optim.Optimizer]] = 0.2,
         reward_normalization: bool = False,
-        ignore_done: bool = False,
         estimation_step: int = 1,
         exploration_noise: Optional[BaseNoise] = None,
         deterministic_eval: bool = True,
         **kwargs: Any,
     ) -> None:
-        super().__init__(None, None, None, None, action_range, tau, gamma,
-                         exploration_noise, reward_normalization, ignore_done,
-                         estimation_step, **kwargs)
+        super().__init__(
+            None, None, None, None, tau, gamma, exploration_noise,
+            reward_normalization, estimation_step, **kwargs
+        )
         self.actor, self.actor_optim = actor, actor_optim
         self.critic1, self.critic1_old = critic1, deepcopy(critic1)
         self.critic1_old.eval()
@@ -101,13 +99,9 @@ class SACPolicy(DDPGPolicy):
         return self
 
     def sync_weight(self) -> None:
-        for o, n in zip(
-            self.critic1_old.parameters(), self.critic1.parameters()
-        ):
+        for o, n in zip(self.critic1_old.parameters(), self.critic1.parameters()):
             o.data.copy_(o.data * (1.0 - self._tau) + n.data * self._tau)
-        for o, n in zip(
-            self.critic2_old.parameters(), self.critic2.parameters()
-        ):
+        for o, n in zip(self.critic2_old.parameters(), self.critic2.parameters()):
             o.data.copy_(o.data * (1.0 - self._tau) + n.data * self._tau)
 
     def forward(  # type: ignore
@@ -122,54 +116,45 @@ class SACPolicy(DDPGPolicy):
         assert isinstance(logits, tuple)
         dist = Independent(Normal(*logits), 1)
         if self._deterministic_eval and not self.training:
-            x = logits[0]
+            act = logits[0]
         else:
-            x = dist.rsample()
-        y = torch.tanh(x)
-        act = y * self._action_scale + self._action_bias
-        y = self._action_scale * (1 - y.pow(2)) + self.__eps
-        log_prob = dist.log_prob(x).unsqueeze(-1)
-        log_prob = log_prob - torch.log(y).sum(-1, keepdim=True)
-        if self._noise is not None and self.training and not self.updating:
-            act += to_torch_as(self._noise(act.shape), act)
-        act = act.clamp(self._range[0], self._range[1])
+            act = dist.rsample()
+        log_prob = dist.log_prob(act).unsqueeze(-1)
+        # apply correction for Tanh squashing when computing logprob from Gaussian
+        # You can check out the original SAC paper (arXiv 1801.01290): Eq 21.
+        # in appendix C to get some understanding of this equation.
+        if self.action_scaling and self.action_space is not None:
+            action_scale = to_torch_as(
+                (self.action_space.high - self.action_space.low) / 2.0, act
+            )
+        else:
+            action_scale = 1.0  # type: ignore
+        squashed_action = torch.tanh(act)
+        log_prob = log_prob - torch.log(
+            action_scale * (1 - squashed_action.pow(2)) + self.__eps
+        ).sum(-1, keepdim=True)
         return Batch(
-            logits=logits, act=act, state=h, dist=dist, log_prob=log_prob)
+            logits=logits, act=squashed_action, state=h, dist=dist, log_prob=log_prob
+        )
 
-    def _target_q(
-        self, buffer: ReplayBuffer, indice: np.ndarray
-    ) -> torch.Tensor:
-        batch = buffer[indice]  # batch.obs: s_{t+n}
-        with torch.no_grad():
-            obs_next_result = self(batch, input='obs_next')
-            a_ = obs_next_result.act
-            target_q = torch.min(
-                self.critic1_old(batch.obs_next, a_),
-                self.critic2_old(batch.obs_next, a_),
-            ) - self._alpha * obs_next_result.log_prob
+    def _target_q(self, buffer: ReplayBuffer, indices: np.ndarray) -> torch.Tensor:
+        batch = buffer[indices]  # batch.obs: s_{t+n}
+        obs_next_result = self(batch, input='obs_next')
+        a_ = obs_next_result.act
+        target_q = torch.min(
+            self.critic1_old(batch.obs_next, a_),
+            self.critic2_old(batch.obs_next, a_),
+        ) - self._alpha * obs_next_result.log_prob
         return target_q
 
     def learn(self, batch: Batch, **kwargs: Any) -> Dict[str, float]:
-        weight = batch.pop("weight", 1.0)
-
-        # critic 1
-        current_q1 = self.critic1(batch.obs, batch.act).flatten()
-        target_q = batch.returns.flatten()
-        td1 = current_q1 - target_q
-        critic1_loss = (td1.pow(2) * weight).mean()
-        # critic1_loss = F.mse_loss(current_q1, target_q)
-        self.critic1_optim.zero_grad()
-        critic1_loss.backward()
-        self.critic1_optim.step()
-
-        # critic 2
-        current_q2 = self.critic2(batch.obs, batch.act).flatten()
-        td2 = current_q2 - target_q
-        critic2_loss = (td2.pow(2) * weight).mean()
-        # critic2_loss = F.mse_loss(current_q2, target_q)
-        self.critic2_optim.zero_grad()
-        critic2_loss.backward()
-        self.critic2_optim.step()
+        # critic 1&2
+        td1, critic1_loss = self._mse_optimizer(
+            batch, self.critic1, self.critic1_optim
+        )
+        td2, critic2_loss = self._mse_optimizer(
+            batch, self.critic2, self.critic2_optim
+        )
         batch.weight = (td1 + td2) / 2.0  # prio-buffer
 
         # actor
@@ -177,8 +162,10 @@ class SACPolicy(DDPGPolicy):
         a = obs_result.act
         current_q1a = self.critic1(batch.obs, a).flatten()
         current_q2a = self.critic2(batch.obs, a).flatten()
-        actor_loss = (self._alpha * obs_result.log_prob.flatten()
-                      - torch.min(current_q1a, current_q2a)).mean()
+        actor_loss = (
+            self._alpha * obs_result.log_prob.flatten() -
+            torch.min(current_q1a, current_q2a)
+        ).mean()
         self.actor_optim.zero_grad()
         actor_loss.backward()
         self.actor_optim.step()
