@@ -1,7 +1,8 @@
-import torch
+from typing import Any, Dict, List, Optional, Sequence, Tuple, Type, Union
+
 import numpy as np
+import torch
 from torch import nn
-from typing import Any, Dict, List, Type, Tuple, Union, Optional, Sequence
 
 ModuleType = Type[nn.Module]
 
@@ -11,10 +12,11 @@ def miniblock(
     output_size: int = 0,
     norm_layer: Optional[ModuleType] = None,
     activation: Optional[ModuleType] = None,
+    linear_layer: Type[nn.Linear] = nn.Linear,
 ) -> List[nn.Module]:
     """Construct a miniblock with given input/output-size, norm layer and \
     activation."""
-    layers: List[nn.Module] = [nn.Linear(input_size, output_size)]
+    layers: List[nn.Module] = [linear_layer(input_size, output_size)]
     if norm_layer is not None:
         layers += [norm_layer(output_size)]  # type: ignore
     if activation is not None:
@@ -31,17 +33,19 @@ class MLP(nn.Module):
     :param int input_dim: dimension of the input vector.
     :param int output_dim: dimension of the output vector. If set to 0, there
         is no final linear layer.
-    :param hidden_sizes: shape of MLP passed in as a list, not incluing
+    :param hidden_sizes: shape of MLP passed in as a list, not including
         input_dim and output_dim.
     :param norm_layer: use which normalization before activation, e.g.,
-        ``nn.LayerNorm`` and ``nn.BatchNorm1d``, defaults to no normalization.
+        ``nn.LayerNorm`` and ``nn.BatchNorm1d``. Default to no normalization.
         You can also pass a list of normalization modules with the same length
         of hidden_sizes, to use different normalization module in different
         layers. Default to no normalization.
     :param activation: which activation to use after each layer, can be both
-        the same actvition for all layers if passed in nn.Module, or different
+        the same activation for all layers if passed in nn.Module, or different
         activation for different Modules if passed in a list. Default to
         nn.ReLU.
+    :param device: which device to create this model on. Default to None.
+    :param linear_layer: use this module as linear layer. Default to nn.Linear.
     """
 
     def __init__(
@@ -50,9 +54,9 @@ class MLP(nn.Module):
         output_dim: int = 0,
         hidden_sizes: Sequence[int] = (),
         norm_layer: Optional[Union[ModuleType, Sequence[ModuleType]]] = None,
-        activation: Optional[Union[ModuleType, Sequence[ModuleType]]]
-        = nn.ReLU,
+        activation: Optional[Union[ModuleType, Sequence[ModuleType]]] = nn.ReLU,
         device: Optional[Union[str, int, torch.device]] = None,
+        linear_layer: Type[nn.Linear] = nn.Linear,
     ) -> None:
         super().__init__()
         self.device = device
@@ -61,8 +65,7 @@ class MLP(nn.Module):
                 assert len(norm_layer) == len(hidden_sizes)
                 norm_layer_list = norm_layer
             else:
-                norm_layer_list = [
-                    norm_layer for _ in range(len(hidden_sizes))]
+                norm_layer_list = [norm_layer for _ in range(len(hidden_sizes))]
         else:
             norm_layer_list = [None] * len(hidden_sizes)
         if activation:
@@ -70,27 +73,28 @@ class MLP(nn.Module):
                 assert len(activation) == len(hidden_sizes)
                 activation_list = activation
             else:
-                activation_list = [
-                    activation for _ in range(len(hidden_sizes))]
+                activation_list = [activation for _ in range(len(hidden_sizes))]
         else:
             activation_list = [None] * len(hidden_sizes)
         hidden_sizes = [input_dim] + list(hidden_sizes)
         model = []
         for in_dim, out_dim, norm, activ in zip(
-                hidden_sizes[:-1], hidden_sizes[1:],
-                norm_layer_list, activation_list):
-            model += miniblock(in_dim, out_dim, norm, activ)
+            hidden_sizes[:-1], hidden_sizes[1:], norm_layer_list, activation_list
+        ):
+            model += miniblock(in_dim, out_dim, norm, activ, linear_layer)
         if output_dim > 0:
-            model += [nn.Linear(hidden_sizes[-1], output_dim)]
+            model += [linear_layer(hidden_sizes[-1], output_dim)]
         self.output_dim = output_dim or hidden_sizes[-1]
         self.model = nn.Sequential(*model)
 
-    def forward(
-        self, x: Union[np.ndarray, torch.Tensor]
-    ) -> torch.Tensor:
-        x = torch.as_tensor(
-            x, device=self.device, dtype=torch.float32)  # type: ignore
-        return self.model(x.flatten(1))
+    def forward(self, s: Union[np.ndarray, torch.Tensor]) -> torch.Tensor:
+        if self.device is not None:
+            s = torch.as_tensor(
+                s,
+                device=self.device,  # type: ignore
+                dtype=torch.float32,
+            )
+        return self.model(s.flatten(1))  # type: ignore
 
 
 class Net(nn.Module):
@@ -103,12 +107,12 @@ class Net(nn.Module):
     :param action_shape: int or a sequence of int of the shape of action.
     :param hidden_sizes: shape of MLP passed in as a list.
     :param norm_layer: use which normalization before activation, e.g.,
-        ``nn.LayerNorm`` and ``nn.BatchNorm1d``, defaults to no normalization.
+        ``nn.LayerNorm`` and ``nn.BatchNorm1d``. Default to no normalization.
         You can also pass a list of normalization modules with the same length
         of hidden_sizes, to use different normalization module in different
         layers. Default to no normalization.
     :param activation: which activation to use after each layer, can be both
-        the same actvition for all layers if passed in nn.Module, or different
+        the same activation for all layers if passed in nn.Module, or different
         activation for different Modules if passed in a list. Default to
         nn.ReLU.
     :param device: specify the device when the network actually runs. Default
@@ -118,13 +122,13 @@ class Net(nn.Module):
     :param bool concat: whether the input shape is concatenated by state_shape
         and action_shape. If it is True, ``action_shape`` is not the output
         shape, but affects the input shape only.
-    :param int num_atoms: in order to expand to the net of distributional RL,
-         defaults to 1 (not use).
+    :param int num_atoms: in order to expand to the net of distributional RL.
+        Default to 1 (not use).
     :param bool dueling_param: whether to use dueling network to calculate Q
         values (for Dueling DQN). If you want to use dueling option, you should
         pass a tuple of two dict (first for Q and second for V) stating
         self-defined arguments as stated in
-        class:`~tianshou.utils.net.common.MLP`. Defaults to None.
+        class:`~tianshou.utils.net.common.MLP`. Default to None.
 
     .. seealso::
 
@@ -139,7 +143,7 @@ class Net(nn.Module):
     def __init__(
         self,
         state_shape: Union[int, Sequence[int]],
-        action_shape: Optional[Union[int, Sequence[int]]] = 0,
+        action_shape: Union[int, Sequence[int]] = 0,
         hidden_sizes: Sequence[int] = (),
         norm_layer: Optional[ModuleType] = None,
         activation: Optional[ModuleType] = nn.ReLU,
@@ -153,14 +157,15 @@ class Net(nn.Module):
         self.device = device
         self.softmax = softmax
         self.num_atoms = num_atoms
-        input_dim = np.prod(state_shape)
-        action_dim = np.prod(action_shape) * num_atoms
+        input_dim = int(np.prod(state_shape))
+        action_dim = int(np.prod(action_shape)) * num_atoms
         if concat:
             input_dim += action_dim
         self.use_dueling = dueling_param is not None
         output_dim = action_dim if not self.use_dueling and not concat else 0
-        self.model = MLP(input_dim, output_dim, hidden_sizes,
-                         norm_layer, activation, device)
+        self.model = MLP(
+            input_dim, output_dim, hidden_sizes, norm_layer, activation, device
+        )
         self.output_dim = self.model.output_dim
         if self.use_dueling:  # dueling DQN
             q_kwargs, v_kwargs = dueling_param  # type: ignore
@@ -169,17 +174,21 @@ class Net(nn.Module):
                 q_output_dim, v_output_dim = action_dim, num_atoms
             q_kwargs: Dict[str, Any] = {
                 **q_kwargs, "input_dim": self.output_dim,
-                "output_dim": q_output_dim}
+                "output_dim": q_output_dim,
+                "device": self.device
+            }
             v_kwargs: Dict[str, Any] = {
                 **v_kwargs, "input_dim": self.output_dim,
-                "output_dim": v_output_dim}
+                "output_dim": v_output_dim,
+                "device": self.device
+            }
             self.Q, self.V = MLP(**q_kwargs), MLP(**v_kwargs)
             self.output_dim = self.Q.output_dim
 
     def forward(
         self,
         s: Union[np.ndarray, torch.Tensor],
-        state: Optional[Any] = None,
+        state: Any = None,
         info: Dict[str, Any] = {},
     ) -> Tuple[torch.Tensor, Any]:
         """Mapping: s -> flatten (inside MLP)-> logits."""
@@ -221,8 +230,8 @@ class Recurrent(nn.Module):
             num_layers=layer_num,
             batch_first=True,
         )
-        self.fc1 = nn.Linear(np.prod(state_shape), hidden_layer_size)
-        self.fc2 = nn.Linear(hidden_layer_size, np.prod(action_shape))
+        self.fc1 = nn.Linear(int(np.prod(state_shape)), hidden_layer_size)
+        self.fc2 = nn.Linear(hidden_layer_size, int(np.prod(action_shape)))
 
     def forward(
         self,
@@ -236,8 +245,7 @@ class Recurrent(nn.Module):
         training mode, s should be with shape ``[bsz, len, dim]``. See the code
         and comment for more detail.
         """
-        s = torch.as_tensor(
-            s, device=self.device, dtype=torch.float32)  # type: ignore
+        s = torch.as_tensor(s, device=self.device, dtype=torch.float32)  # type: ignore
         # s [bsz, len, dim] (training) or [bsz, dim] (evaluation)
         # In short, the tensor's shape in training phase is longer than which
         # in evaluation phase.
@@ -250,9 +258,49 @@ class Recurrent(nn.Module):
         else:
             # we store the stack data in [bsz, len, ...] format
             # but pytorch rnn needs [len, bsz, ...]
-            s, (h, c) = self.nn(s, (state["h"].transpose(0, 1).contiguous(),
-                                    state["c"].transpose(0, 1).contiguous()))
+            s, (h, c) = self.nn(
+                s, (
+                    state["h"].transpose(0, 1).contiguous(),
+                    state["c"].transpose(0, 1).contiguous()
+                )
+            )
         s = self.fc2(s[:, -1])
         # please ensure the first dim is batch size: [bsz, len, ...]
-        return s, {"h": h.transpose(0, 1).detach(),
-                   "c": c.transpose(0, 1).detach()}
+        return s, {"h": h.transpose(0, 1).detach(), "c": c.transpose(0, 1).detach()}
+
+
+class ActorCritic(nn.Module):
+    """An actor-critic network for parsing parameters.
+
+    Using ``actor_critic.parameters()`` instead of set.union or list+list to avoid
+    issue #449.
+
+    :param nn.Module actor: the actor network.
+    :param nn.Module critic: the critic network.
+    """
+
+    def __init__(self, actor: nn.Module, critic: nn.Module) -> None:
+        super().__init__()
+        self.actor = actor
+        self.critic = critic
+
+
+class DataParallelNet(nn.Module):
+    """DataParallel wrapper for training agent with multi-GPU.
+
+    This class does only the conversion of input data type, from numpy array to torch's
+    Tensor. If the input is a nested dictionary, the user should create a similar class
+    to do the same thing.
+
+    :param nn.Module net: the network to be distributed in different GPUs.
+    """
+
+    def __init__(self, net: nn.Module) -> None:
+        super().__init__()
+        self.net = nn.DataParallel(net)
+
+    def forward(self, s: Union[np.ndarray, torch.Tensor], *args: Any,
+                **kwargs: Any) -> Tuple[Any, Any]:
+        if not isinstance(s, torch.Tensor):
+            s = torch.as_tensor(s, dtype=torch.float32)
+        return self.net(s=s.cuda(), *args, **kwargs)
